import sys
import string


try:
    from numpy import *
except ImportError:
    print "Please install Numerical Python (NumPy) version 1.3 or later"
    sys.exit(2)

class Mapping(list): pass

class sparsearray:
    def __init__(self, index):
        index = set(index)
        index = list(index)
        index.sort()
        n = len(index)
        self.index = array(index, int)
        self.value = zeros(n)
    def __getitem__(self, key):
        if type(key)==slice:
            start, stop, step = key.start, key.stop, key.step
            i = self.index.searchsorted(start)
            j = i + self.index[i:].searchsorted(stop)
            return self.value[i:j]
        else:
            i = self.index.searchsorted(key)
            if i==len(self.index) or self.index[i]!=key:
                return 0.0
            else:
                return self.value[i]
    def __setitem__(self, key, value):
        n = len(self.index)
        if n==0:
            return
        elif type(key)==slice:
            start, stop, step = key.start, key.stop, key.step
            i = self.index.searchsorted(start)
            j = i + self.index[i:].searchsorted(stop)
            self.value[i:j] = value
        else:
            i = self.index.searchsorted(key)
            if i > len(self.index):
                raise IndexError, "index out of bounds"
            self.value[i] = value

def parse_cigar(cigar):
    result = []
    number = ""
    for c in cigar:
        if c in "0123456789":
            number += c
        elif c in 'MIDNSHP':
            number = int(number)
            if c=='M':
                result.append(('M', number))
            elif c=='I':
                result.append(('I', number))
            elif c=='S':
                result.append(('S', number))
            elif c in 'DN':
                result.append(('D', number))
            number = ''
        else:
            raise Exception, "Unknown code in cigar line: %s" % cigar
    return result

def parse_md(md):
    md = md.replace("MD:Z:", "")
    result = []
    number = ""
    deletion = ""
    for c in md:
        if c in "0123456789":
            if deletion:
                result.append(deletion)
                deletion = ""
                number = ""
            number += c
        elif c=='^':
            result.append(number)
            deletion = "^"
            number = ""
        elif deletion:
            deletion += c
        else:
            if number!="":
                result.append(number)
            result.append(c)
            number = ""
    assert deletion==""
    if number!="":
        result.append(number)
    return result

def parse_number(field):
    tag, vtype, value = field.split(":")
    if vtype=='i':
        return int(value)
    elif vtype=='f':
        return float(value)

def create_alignment(cigar, md, start, strand):
    n = sum([count for operation, count in cigar])
    i = 0
    status = zeros(n, int)
    for operation, count in cigar:
        if operation=='M':
            pass
        elif operation in 'IS':
            status[i:i+count] = 3
        elif operation=='D':
            status[i:i+count] = 2
        else:
            raise Exception, "Unknown operation in cigar line: %s" % operation
        i += count
    i = 0
    for operation in md:
        if operation[0] in "0123456789":
            m = int(operation)
            j = 0
            while j < m:
                if status[i]!=3:
                    j += 1
                i += 1
            while i < n and status[i]==3:
                i += 1
        elif operation.startswith('^'):
            deletion = operation[1:]
            m = len(deletion)
            i += m
        else:
            status[i] = 1 # operation contains the mismatched nucleotide
            i += 1
    assert i==n
    position = zeros(n, int)
    i = 0
    if strand=='+':
        j = start
        for operation, count in cigar:
            if operation in 'MD':
                position[i:i+count] = arange(j,j+count)
                j += count
            elif operation in 'IS':
                position[i:i+count] = j
            else:
                raise Exception, "Unknown operation in cigar line: %s" % operation
            i += count
        if operation=='S':
            position[-count:] -= 1
    elif strand=='-':
        status = status[::-1]
        cigar = cigar[::-1]
        j = 0
        for operation, count in cigar:
            if operation in 'MD':
                position[i:i+count] = arange(j,j-count,-1)
                j -= count
            elif operation in 'IS':
                position[i:i+count] = j
            else:
                raise Exception, "Unknown operation in cigar line: %s" % operation
            i += count
        if operation=='S':
            start -= count
        position += start-position[-1]
    return position, status

translation = string.maketrans("ACGT", "TGCA")

def reverse_complement(sequence):
    return sequence[::-1].translate(translation)

def read_mappings(handle):
    mappings = {}
    for line in handle:
        if line.startswith("@"):
            continue
        words = line.split()
        flag = int(words[1])
        if flag & 16:
            strand = '-'
        else:
            strand = '+'
        chromosome = words[2]
        start = int(words[3])-1
        cigar = parse_cigar(words[5])
        tag = words[9]
        weight = None
        count = None
        for word in words[10:]:
            if word.startswith("MD:Z:"):
                md = parse_md(word)
            elif word.startswith("XW:f:"):
                weight = parse_number(word)
            elif word.startswith("XC:"):
                count = parse_number(word)
        if count==None:
            print "XC:i: field is missing in the input file; run python cmc.py --help for more information"
            sys.exit(2)
        if strand=='-':
            tag = reverse_complement(tag)
        if not tag in mappings:
            mappings[tag] = Mapping()
            mappings[tag].weight = []
            mappings[tag].count = float(count)
        else:
            assert mappings[tag].count == float(count)
        position, status = create_alignment(cigar, md, start, strand)
        if strand=='+':
            end = position[-1]+1
        elif strand=='-':
            end = position[0]+1
        mappings[tag].append([chromosome, strand, start, end, status, position])
        if weight!=None:
            mappings[tag].weight.append(weight)
    for tag in mappings:
        if mappings[tag].weight:
            mappings[tag].weight = array(mappings[tag].weight)
        else:
            n = len(mappings[tag])
            mappings[tag].weight = ones(n)/n
    return mappings

def create_prior(mappings):
    MAXSIZE = 10000
    prior = {}
    for tag in mappings:
        for row in mappings[tag]:
            [chromosome, strand, start, end] = row[:4]
            if not chromosome in prior:
                prior[chromosome] = {'+': [], '-': []}
            prior[chromosome][strand].extend(range(start, end))
            n = len(prior[chromosome][strand])
            if n > MAXSIZE:
                prior[chromosome][strand] = list(set(prior[chromosome][strand]))
                if len(prior[chromosome][strand])==MAXSIZE:
                    raise Exception, "failed to reduce memory"
                MAXSIZE = int(1.2*n)
    for chromosome in prior:
        for strand in '+-':
            prior[chromosome][strand] = sparsearray(prior[chromosome][strand])
    return prior

def choose(mappings, prior, profile):
    ends = (len(profile)-1)/2
    logprofile = log(profile)
    maxchange = 0.0
    totalchange = 0.0
    totaltags = 0.0
    for tag in mappings:
        m = mappings[tag]
        n = len(m)
        oldweight = array(m.weight[:])
        m.weight[:] = 0.0
        for i in range(n):
            [chromosome, strand, start, end] = m[i][:4]
            m.weight[i] = prior[chromosome][strand][start:end].mean()
        total = sum(m.weight)
        pseudocounts = sqrt(total)
        m.weight += pseudocounts / n
        m.weight /= (total + pseudocounts)
        m.weight = log(m.weight)
        for i in range(n):
            status = array(m[i][4], dtype=bool)
            factor = dot(logprofile[:ends], status[:ends])
            factor += dot(logprofile[-ends:], status[-ends:])
            factor += logprofile[ends] * sum(status[ends:-ends])
            m.weight[i] += factor
        m.weight -= mean(m.weight) # to avoid roundoff errors
        m.weight = exp(m.weight)
        total = sum(m.weight)
        m.weight /= total
        weight_difference = abs(m.weight-oldweight)
        maxchange = max(max(weight_difference), maxchange)
        totalchange += 0.5*sum(weight_difference)*mappings[tag].count
        totaltags += mappings[tag].count
    return maxchange, totalchange, totaltags

def calculate_prior(mappings, prior):
    for chromosome in prior:
        for strand in prior[chromosome]:
            prior[chromosome][strand][:] = 0.0
    for tag in mappings:
        m = mappings[tag]
        value = float(m.count)
        n = len(m)
        for i in range(n):
            weight = m.weight[i]
            [chromosome, strand, start, end] = m[i][:4]
            row = prior[chromosome][strand][start:end]
            if value*weight < 0: 
                print tag, i, value, weight
                raise Exception
            row += value*weight
            # This is more efficient than
            # prior[chromosome][strand][start:end] += value*weight

def calculate_profile(mappings, profile):
    ends = (len(profile)-1)/2
    profile[:] = 0.0
    total = zeros(len(profile))
    for tag in mappings:
        m = mappings[tag]
        count = m.count
        n = len(m)
        for i in range(n):
            [chromosome, strand, start, end] = m[i][:4]
            status = array(m[i][4], dtype=bool)
            value = count * m.weight[i]
            profile[:ends] += status[:ends] * value
            profile[-ends:] += status[-ends:] * value
            profile[ends] += sum(status[ends:-ends]) * value
            total[:ends] += value
            total[-ends:] += value
            total[ends] += (len(status)-2*ends) * value
    pseudocounts = sqrt(total)
    profile[:] += pseudocounts
    total += pseudocounts
    profile[:] /= total

def save(input, output, mappings):
    d = {}
    for line in input:
        if line.startswith("@"):
            continue
        words = line.split()
        flag = int(words[1])
        if flag & 16:
            strand = '-'
        else:
            strand = '+'
        tag = words[9]
        if strand=='-':
            tag = reverse_complement(tag)
        n = len(words)
        for j in range(10,n):
            word = words[j]
            if word.startswith("XW:f:"):
                break
        else:
            j = n
            words.append('')
        if tag in mappings:
            if not tag in d:
                d[tag] = 0
            k = d[tag]
            d[tag] += 1
            weight = mappings[tag].weight[k]
            word = "XW:f:%f" % weight
            words[j] = word
            line = "\t".join(words) + "\n"
        output.write(line)
    output.close()
    input.close()

def version():
    print """\
Cross-mapping correction version 2010.03.15
Copyright (C) 2009-2010 Michiel JL de Hoon (mdehoon@gsc.riken.jp),
RIKEN Omics Science Center, Yokohama, Japan.
    """

def usage():
    version()
    print """\
Usage:

python cmc.py -i <input file> -o <output file> [-n iterations] [-e ends]

     Options:
        -i <input file>    name of the input file in SAM format
        -o <output file>   name of the output file
        -n iterations      number of iterations to run (default: 1)
        -e ends            number of nucleotides at the ends of the alignment
                           for which a separate error rate is calculated
                           (default: 3)

The input SAM file should have the MD:s: field to show any mismatches in the
alignment; see the SAM file format description for more information. This
field can be generated using the samtools software.

In addition, each line in the SAM file should have an XC:i:<number> field,
where <number> is the count of the sequence (i.e., how often it appeared in
the library).

Optionally, each line can contain a XW:f:<weight> field, where the weight is
the initial weight to be used in the iteration.
"""

def main(argv):
    import getopt
    iterations = 1
    ends = 3
    try:
        opts, args = getopt.getopt(argv, "vhi:o:n:e:", ["--version", "--help"])
    except getopt.GetoptError, error:
        print str(error)
        usage()
        sys.exit(2)
    input_filename = None
    output_filename = None
    for opt, arg in opts:
        if opt in ("-h", "--help"):
            usage()
            sys.exit()
        elif opt in ("-v", "--version"):
            version()
            sys.exit()
        elif opt=="-i":
            input_filename = arg
        elif opt=="-o":
            output_filename = arg
        elif opt=="-n":
            iterations = int(arg)
        elif opt=="-e":
            ends = int(arg)
    if not input_filename or not output_filename:
        usage()
        sys.exit(2)
    try:
        input = open(input_filename)
    except IOError:
        print "Failed to open input file", input_filename
        sys.exit(2)
    mappings = read_mappings(input)
    input.close()
    profile = ones(2*ends+1)
    prior = create_prior(mappings)
    for iteration in range(iterations):
        print "Running iteration %d of %d" % (iteration+1, iterations),
        calculate_prior(mappings, prior)
        calculate_profile(mappings, profile)
        maxchange, totalchange, total = choose(mappings, prior, profile)
        print "max change: %5.3f, relative change: %5.3g" % (maxchange, totalchange / total)
        # The "change" variable can be used to assess the convergence rate
    try:
        input = open(input_filename)
    except IOError:
        print "Failed to open input file", input_filename, "when saving result"
        sys.exit(2)
    try:
        output = open(output_filename, 'w')
    except IOError:
        print "Failed to open output file", output_filename
        sys.exit(2)
    print "Iteration finished; saving results to", output_filename
    save(input, output, mappings)
    input.close()
    output.close()

if __name__ == "__main__":
    main(sys.argv[1:])
